#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#
import json
from dataclasses import dataclass
from enum import Enum
from typing import Any


@dataclass
class EventMetric:
    """
    A metric attached to an event. It can have a single value.
    It is understood that the timestamp of the event is the timestamp of the metric. In OpenTelemetry,
    these metrics are reported as event attributes. For other backends they are serialized
    as a JSON object in the event. For spans they are normally attached to the span end event.
    """

    name: str
    value: Any


class EventMetrics(dict[str, EventMetric]):
    """A set of metrics for an event."""

    def add_metric(self, name: str, value: Any) -> EventMetric:
        """Add a new metric to the set of metrics."""
        self[name] = EventMetric(name, value)
        return self[name]

    def to_json(self) -> dict[str, Any]:
        """Convert all metrics to a JSON-compatible dictionary.

        Returns:
            dict: JSON-serializable representation of all metrics
        """
        return {
            metric.name: json.dumps(metric.value) if isinstance(metric.value, str) else metric.value
            for metric in self.values()
        }

    @classmethod
    def from_json(cls, data: dict) -> "EventMetrics":
        """Create a EventMetrics instance from a JSON-compatible dictionary.

        Args:
            data: Dictionary containing the metrics data

        Returns:
            EventMetrics: New metrics collection created from the data
        """
        metrics = cls()

        def _decode_value(value: Any) -> Any:
            try:
                if isinstance(value, str) and value.startswith('"'):
                    return json.loads(value)
                else:
                    return value
            except (json.JSONDecodeError, TypeError):
                return value

        for name, value in data.items():
            metrics.add_metric(name, _decode_value(value))

        return metrics

    @classmethod
    def merge(cls, metrics: "EventMetrics", other: "EventMetrics") -> "EventMetrics":
        """Merge two sets of metrics."""
        merged = cls()
        for metric in metrics.values():
            merged.add_metric(metric.name, metric.value)
        for metric in other.values():
            merged.add_metric(metric.name, metric.value)
        return merged


class ApplicationMetrics(EventMetrics):
    """A set of metrics for an application."""

    @classmethod
    def create(
        cls,
        rank: int | None = None,
        world_size: int | None = None,
        node_name: str | None = None,
        timezone: str | None = None,
        total_iterations: int | None = None,
        checkpoint_enabled: bool | None = None,
        checkpoint_strategy: str | None = None,
    ) -> "ApplicationMetrics":
        metrics = cls()
        if rank is not None:
            metrics.add_metric("rank", rank)
        if world_size is not None:
            metrics.add_metric("world_size", world_size)
        if node_name is not None:
            metrics.add_metric("node_name", node_name)
        if timezone is not None:
            metrics.add_metric("timezone", timezone)
        if total_iterations is not None:
            metrics.add_metric("total_iterations", total_iterations)
        if checkpoint_enabled is not None:
            metrics.add_metric("checkpoint_enabled", checkpoint_enabled)
        if checkpoint_strategy is not None:
            metrics.add_metric("checkpoint_strategy", checkpoint_strategy)
        return metrics


class IterationMetrics(EventMetrics):
    """
    A set of metrics for one or more iterations, either for model training, validation, or testing.
    For testing and validation, these are normally used to report the metrics at the end of the loop, or span.
    For training, these are normally used to report the metrics using the TRAINING_ITERATIONS event.
    """

    @classmethod
    def create(
        cls,
        current_iteration: int | None = None,
        num_iterations: int | None = None,
        interval: int | None = None,
        average_iteration_time: float | None = None,
        average_forward_time: float | None = None,
        average_backward_time: float | None = None,
        average_dataloader_time: float | None = None,
        tflops: float | None = None,
        tokens_per_second: float | None = None,
        loss: float | None = None,
        batch_size: int | None = None,
    ) -> "IterationMetrics":
        """
        Create a IterationMetrics instance.

        Args:
            current_iteration: The current iteration number of the loop, the last one if reported at the end of the loop
            num_iterations: The total number of iterations in the loop or since last reporting in the case of TRAINING_ITERATIONS
            interval: The interval between the current and previous iteration, where a similar event or span was reported. Note
            that for TRAINING_ITERATIONS, this is normally the same as num_iterations.
            average_iteration_time: The average time per iteration
            average_forward_time: The average model forward time per iteration
            average_backward_time: The average model backward time per iteration
            average_dataloader_time: The average dataloader time per iteration
            tflops: The number of tera-floating point operations per second for each iteration
            tokens_per_second: The number of tokens processed per second for each iteration, needed if tflops cannot be calculated
            loss: The current loss value
            batch_size: The number of samples or tokens processed per iteration, also known as the batch size
        """
        metrics = cls()
        if current_iteration is not None:
            metrics.add_metric("current_iteration", current_iteration)
        if num_iterations is not None:
            metrics.add_metric("num_iterations", num_iterations)
        if interval is not None:
            metrics.add_metric("interval", interval)
        if average_iteration_time is not None:
            metrics.add_metric("average_iteration_time", average_iteration_time)
        if average_forward_time is not None:
            metrics.add_metric("average_forward_time", average_forward_time)
        if average_backward_time is not None:
            metrics.add_metric("average_backward_time", average_backward_time)
        if average_dataloader_time is not None:
            metrics.add_metric("average_dataloader_time", average_dataloader_time)
        if tflops is not None:
            metrics.add_metric("tflops", tflops)
        if tokens_per_second is not None:
            metrics.add_metric("tokens_per_second", tokens_per_second)
        if loss is not None:
            metrics.add_metric("loss", loss)
        if batch_size is not None:
            metrics.add_metric("batch_size", batch_size)
        return metrics


class CheckPointType(str, Enum):
    """The type of checkpoint, global or local."""

    # Global checkpoint, saved to remote storage, and persistent.
    GLOBAL = "global"
    # Local checkpoint, saved to local storage, and non-persistent.
    LOCAL = "local"


class CheckpointMetrics(EventMetrics):
    """
    A set of metrics for a checkpoint save or load events.

    The checkpoint type is either global or local.
    The current iteration is the iteration number at the checkpoint save or load.
    The number of iterations since the previous checkpoint save, not needed for checkpoint load.
    The interval is the number of iterations between the current and previous checkpoint save,
    normally the same as num_iterations unless some iterations were excluded from checkpointing.
    The checkpoint size is the size of the checkpoint in bytes.
    The checkpoint directory is the directory where the checkpoint is saved or loaded from.
    """

    @classmethod
    def create(
        cls,
        checkpoint_type: CheckPointType | None = None,
        current_iteration: int | None = None,
        num_iterations: int | None = None,
        interval: int | None = None,
        checkpoint_size: int | None = None,
        checkpoint_directory: str | None = None,
    ) -> "CheckpointMetrics":
        metrics = cls()
        if checkpoint_type is not None:
            metrics.add_metric("checkpoint_type", checkpoint_type)
        if current_iteration is not None:
            metrics.add_metric("current_iteration", current_iteration)
        if num_iterations is not None:
            metrics.add_metric("num_iterations", num_iterations)
        if interval is not None:
            metrics.add_metric("interval", interval)
        if checkpoint_size is not None:
            metrics.add_metric("checkpoint_size", checkpoint_size)
        if checkpoint_directory is not None:
            metrics.add_metric("checkpoint_directory", checkpoint_directory)
        return metrics



#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#

import logging
import os
from dataclasses import fields
from typing import Optional, Type, TypeVar

import yaml  # type: ignore[import-untyped, unused-ignore]

from training_telemetry.config import (
    ApplicationConfig,
    BackendConfig,
    BackendType,
    FileBackendConfig,
    LoggerBackendConfig,
    NVTXBackendConfig,
    OTLPLogsBackendConfig,
    OTLPLogsFormat,
    OTLPTracesBackendConfig,
    TelemetryConfig,
    default_backends,
)

T = TypeVar("T")


def load_config_from_dict(config: dict) -> TelemetryConfig:
    """Create a TelemetryConfig from a dictionary containing some or all of the class properties."""
    if not config:
        return TelemetryConfig()

    config = config.copy()
    config["backends"] = [_backend_from_dict(backend) for backend in config["backends"]] if "backends" in config else []
    if "application" in config:
        config["application"] = _dataclass_from_dict(ApplicationConfig, config["application"])
    if "log_level" in config:
        config["log_level"] = _get_log_level(config["log_level"])
    return TelemetryConfig(**config)


def load_config_from_env(config: dict) -> TelemetryConfig:
    """
    Create a TelemetryConfig from environment variables with the same name as the class properties, but uppercase
    and with the prefix "TRAINING_TELEMETRY_".
    If an environment variable is not set, any value already in the config dictionary, if it exists, will be used.

    Args:
        config: The configuration dictionary to update, for example the config read from file, or set manually from the user.
                Note that this dictionary will be updated in place.
    """
    application: ApplicationConfig = _dataclass_from_env(
        ApplicationConfig,
        config["application"] if "application" in config else {},
        ["TRAINING_TELEMETRY_APPLICATION", "APPLICATION", ""],
    )
    backends = _load_backends_from_env(config, application.rank, application.world_size)
    log_level = _get_log_level(os.getenv("TRAINING_TELEMETRY_LOG_LEVEL", "INFO"))
    log_format = os.getenv("TRAINING_TELEMETRY_LOG_FORMAT")
    return TelemetryConfig(backends=backends, application=application, log_level=log_level, log_format=log_format)


def _load_backends_from_env(config: dict, rank: int, world_size: int) -> list[BackendConfig]:
    """
    Helper function to load_config_from_env to load the backends from environment variables.
    Only load backends that are either in the config dict, or have at least one environment variable set.
    If no backends are loaded, return the default backends.
    """
    backends: list[BackendConfig] = []
    backend_configs_by_type: dict[BackendType, dict] = {}
    for backend_config in config["backends"] if "backends" in config else []:
        backend_configs_by_type[BackendType(backend_config["backend_type"])] = backend_config

    for backend_type in BackendType:
        enabled_in_env = os.getenv(f"TRAINING_TELEMETRY_BACKENDS_{backend_type.value.upper()}_ENABLED_RANKS") in (
            "all",
            str(rank),
            str(rank - world_size),
        )
        enabled_in_config = backend_type in backend_configs_by_type
        if enabled_in_env or enabled_in_config:
            backends.append(
                _backend_from_env(backend_type, backend_configs_by_type[backend_type] if enabled_in_config else {})
            )

    return backends if backends else default_backends()


def load_config(
    config_file: Optional[str | os.PathLike] = None,
    defaults: Optional[dict] = None,
    override_from_env: bool = True,
) -> TelemetryConfig:
    """
    Create a TelemetryConfig from a file and with optional overrides from environment variables, and starting from optional default values.
    The order of precedence is: command line defaults, config file, environment variables. Environment variables will
    override the config file and config file will override command line defaults.

    Args:
        config_file: The path or path-like object to the config file.
        defaults: Optional command-line default set of values that can be passed by the application code.
        override_from_env: If true, the default, the config will be overridden by environment variables, if they are set. These
        environment variables must have the prefix "TRAINING_TELEMETRY_", followed by the name of the class property.
    """
    # If no config arguments are provided, return the default config.
    if not (config_file or defaults or override_from_env):
        return TelemetryConfig()

    if not config_file:
        yaml_config = {}
    else:
        with open(config_file, "r") as file:
            yaml_config = yaml.safe_load(file)
            if not yaml_config:
                raise RuntimeError(f"Could not parse config file {config_file}")
    if defaults:
        config = defaults.copy() | yaml_config
    else:
        config = yaml_config

    if override_from_env:
        return load_config_from_env(config)
    else:
        return load_config_from_dict(config)


def _backend_from_dict(config: dict) -> BackendConfig:
    """Create a BackendConfig from a dictionary by matching the backend type."""
    if isinstance(config["backend_type"], str):
        config["backend_type"] = BackendType(config["backend_type"])

    match config["backend_type"]:
        case BackendType.LOGGER:
            return _dataclass_from_dict(LoggerBackendConfig, config)
        case BackendType.FILE:
            return _dataclass_from_dict(FileBackendConfig, config)
        case BackendType.OTLP_TRACES:
            return _dataclass_from_dict(OTLPTracesBackendConfig, config)
        case BackendType.OTLP_LOGS:
            return _dataclass_from_dict(OTLPLogsBackendConfig, config)
        case BackendType.NVTX:
            return _dataclass_from_dict(NVTXBackendConfig, config)
        case _:
            raise ValueError(f"Unsupported backend type: {config['backend_type']}")


def _backend_from_env(backend_type: BackendType, config: dict) -> BackendConfig:
    """Create a BackendConfig from environment variables by matching the backend type."""
    match backend_type:
        case BackendType.LOGGER:
            return _dataclass_from_env(LoggerBackendConfig, config, ["TRAINING_TELEMETRY_BACKENDS_LOGGER"])
        case BackendType.FILE:
            return _dataclass_from_env(FileBackendConfig, config, ["TRAINING_TELEMETRY_BACKENDS_FILE"])
        case BackendType.OTLP_TRACES:
            return _dataclass_from_env(OTLPTracesBackendConfig, config, ["TRAINING_TELEMETRY_BACKENDS_OTLP_TRACES"])
        case BackendType.OTLP_LOGS:
            return _dataclass_from_env(OTLPLogsBackendConfig, config, ["TRAINING_TELEMETRY_BACKENDS_OTLP_LOGS"])
        case BackendType.NVTX:
            return _dataclass_from_env(NVTXBackendConfig, config, ["TRAINING_TELEMETRY_BACKENDS_NVTX"])
        case _:
            raise ValueError(f"Unsupported backend type: {backend_type}")


def _dataclass_from_env(cls: Type[T], config: dict, env_prefixes: list[str]) -> T:
    """
    Create an instance of a dataclass either from the environment variables, if set, or from the config dictionary.
    If neither environment variables nor the config dictionary have the value for a field, the default value will be used.
    This assumes that the dataclass has default values for all fields, as is the case for the config dataclasses.

    Args:
        config: The configuration dictionary to update, for example the config read from file, or set manually from the user.
                Note that this dictionary will be updated in place.
    """
    for field in fields(cls):  # type: ignore[arg-type]
        env_value = None
        for env_prefix in env_prefixes:
            env_value = os.getenv(f"{env_prefix}_{field.name.upper()}") if env_prefix else os.getenv(field.name.upper())
            if env_value:
                break

        if env_value:
            config[field.name] = env_value
        elif field.name not in config:
            config[field.name] = field.default

    return _dataclass_from_dict(cls, config)


def _dataclass_from_dict(cls: Type[T], config: dict) -> T:
    """
    Create an instance of a dataclass from a dictionary by converting the strings to either
    supported enums, log level, integers or booleans. These are the types supported by
    the config dataclasses.
    """
    for field in fields(cls):  # type: ignore[arg-type]
        if field.name not in config:
            continue

        # Special case for log level because it can be a string or an integer.
        if field.name == "log_level":
            config["log_level"] = _get_log_level(config["log_level"])
            continue

        # The rest is transformation of strings to the supported types.
        if not isinstance(config[field.name], str):
            continue

        if field.type == OTLPLogsFormat:
            config[field.name] = OTLPLogsFormat(config["format"])
        elif field.type == BackendType:
            config[field.name] = BackendType(config["backend_type"])
        elif field.type == bool:
            config[field.name] = config[field.name].upper() == "TRUE"
        elif field.type == int:
            config[field.name] = int(config[field.name])

    return cls(**config)


def _get_log_level(log_level: str | int) -> int:
    """
    Get the log level. If we receive an integer, we return it as is as we assume this is in the form logging.INFO, etc.
    If we receive a string, we assume can be converted to a level. Since python 3.11 there is a getLevelNamesMapping function
    that can be used to get the log level from the name but we need to support python 3.10 too.
    """
    if isinstance(log_level, int):
        return log_level
    if hasattr(logging, "getLevelNamesMapping"):
        level_mapping = getattr(logging, "getLevelNamesMapping")()
        return int(level_mapping[log_level.upper()])
    else:
        return int(getattr(logging, log_level.upper()))



#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#
import os
import tempfile
from pathlib import Path
from typing import Any

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader, TensorDataset
from tzlocal import get_localzone

from training_telemetry.config import TelemetryConfig
from training_telemetry.config_loader import load_config
from training_telemetry.context import checkpoint_save, get_recorder, running, timed_span, training
from training_telemetry.events import Event, EventName
from training_telemetry.metrics import ApplicationMetrics, CheckpointMetrics, CheckPointType, IterationMetrics
from training_telemetry.provider import Provider
from training_telemetry.spans import SpanColor, SpanName
from training_telemetry.torch.utils import get_rank, get_world_size
from training_telemetry.verbosity import Verbosity

# Generate some random data for this example
torch.manual_seed(42)
# Generate random input features and binary labels
X = torch.randn(1000, 10, dtype=torch.float32)
y = (X.sum(dim=1) > 0).float()
dataset = TensorDataset(X, y)
dataloader = DataLoader(dataset, batch_size=32, shuffle=True, num_workers=0)
num_epochs = 100

# Initialize the telemetry provider with a default configuration
config = load_config(
    config_file=Path(__file__).parent / "example_config.yaml",
    defaults={"application": {"job_name": "torch_example", "job_id": "1234567890", "environment": "test"}},
    override_from_env=False,
)
Provider.set_provider(config)


# Define a simple model
class SimpleModel(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.layers = nn.Sequential(nn.Linear(10, 64), nn.ReLU(), nn.Linear(64, 1), nn.Sigmoid())

    def forward(self, x: torch.Tensor) -> Any:
        return self.layers(x)


def get_application_metrics() -> ApplicationMetrics:
    return ApplicationMetrics.create(
        rank=get_rank(),
        world_size=get_world_size(),
        node_name="localhost",
        timezone=str(get_localzone()),
        total_iterations=num_epochs * len(dataloader),
        checkpoint_enabled=True,
        checkpoint_strategy="sync",
    )


@running(metrics=get_application_metrics())
def main() -> None:
    # Initialize model, loss function and optimizer
    model = SimpleModel()
    criterion = nn.BCELoss()
    optimizer = optim.Adam(model.parameters(), lr=0.001, weight_decay=0.0001)

    with training() as training_span:
        current_iteration = 0
        accuracy = torch.tensor(float("nan"))
        loss = torch.tensor(float("nan"))

        for epoch in range(num_epochs):
            for batch_idx, (inputs, targets) in enumerate(dataloader):
                with timed_span(SpanName.ITERATION, color=SpanColor.RED, verbosity=Verbosity.PROFILING):
                    # Forward pass
                    with timed_span(SpanName.MODEL_FORWARD, color=SpanColor.RED, verbosity=Verbosity.PROFILING):
                        outputs = model(inputs)
                        loss = criterion(outputs.squeeze(), targets)

                    # Backward pass and optimize
                    with timed_span(SpanName.ZERO_GRAD, color=SpanColor.GREEN, verbosity=Verbosity.PROFILING):
                        optimizer.zero_grad()
                    with timed_span(SpanName.MODEL_BACKWARD, color=SpanColor.BLUE, verbosity=Verbosity.PROFILING):
                        loss.backward()
                    with timed_span(SpanName.OPTIMIZER_UPDATE, color=SpanColor.YELLOW, verbosity=Verbosity.PROFILING):
                        optimizer.step()

                    # Calculate accuracy
                    predictions = (outputs.squeeze() > 0.5).float()
                    accuracy = (predictions == targets).float().mean()

                    current_iteration += 1

            # Log iteration metrics every 2 epochs
            if epoch % 2 == 0:
                metrics = IterationMetrics.create(
                    current_iteration=current_iteration,
                    num_iterations=len(dataloader),
                    interval=len(dataloader),
                    loss=loss.item(),
                    tflops=100,
                )
                get_recorder().event(Event.create(EventName.TRAINING_ITERATIONS, metrics=metrics), training_span)

            # Save checkpoint every 5 epochs
            if epoch % 5 == 0:
                print(
                    f"Epoch [{epoch+1}/{num_epochs}], "
                    f"Batch [{batch_idx+1}/{len(dataloader)}], "
                    f"Loss: {loss.item():.4f}, "
                    f"Accuracy: {accuracy.item():.4f}"
                )
                checkpoint = {
                    "model_state_dict": model.state_dict(),
                    "optimizer_state_dict": optimizer.state_dict(),
                    "epoch": epoch,
                    "iteration": current_iteration,
                    "loss": loss.item(),
                    "accuracy": accuracy.item(),
                }
                with checkpoint_save() as checkpoint_save_span:
                    with tempfile.TemporaryDirectory() as temp_dir:
                        checkpoint_file_name = os.path.join(temp_dir, f"checkpoint_iter_{current_iteration}.pt")
                        torch.save(checkpoint, checkpoint_file_name)
                        checkpoint_save_span.add_metrics(
                            CheckpointMetrics.create(
                                checkpoint_type=CheckPointType.LOCAL,
                                current_iteration=current_iteration,
                                num_iterations=len(dataloader),
                                checkpoint_directory=temp_dir,
                            )
                        )


if __name__ == "__main__":
    main()



#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#
import uuid
from collections import defaultdict

import nvtx

from training_telemetry.config import BackendType
from training_telemetry.events import Event, ExceptionEvent
from training_telemetry.internal.backend import Backend
from training_telemetry.spans import Span, SpanName
from training_telemetry.utils import get_default_logger
from training_telemetry.verbosity import Verbosity

_logger = get_default_logger()


class NVTXBackend(Backend):
    """Backend implementation that sends telemetry data to NVIDIA Tools Extension (NVTX)."""

    def __init__(self) -> None:
        """Initialize the NVTX backend."""
        self.ranges_by_span_id: dict[uuid.UUID, nvtx.Range] = {}
        self.counts_by_span_name: defaultdict[SpanName, int] = defaultdict(int)

    def type(self) -> BackendType:
        return BackendType.NVTX

    def verbosity(self) -> Verbosity:
        return Verbosity.PROFILING

    def record_start(self, span: Span) -> None:
        """Record the start of a time span."""
        self.counts_by_span_name[span.name] += 1
        name = f"{span.name.value}_{self.counts_by_span_name[span.name]}"
        self.ranges_by_span_id[span.id] = nvtx.start_range(message=name, color=span.color.value if span.color else None)

    def record_stop(self, span: Span) -> None:
        """Record the end of a time span."""
        try:
            range_id = self.ranges_by_span_id.pop(span.id)
            if range_id:
                nvtx.end_range(range_id)
        except KeyError:
            _logger.warning(f"Span {span.id} not found in NVTX ranges")

    def record_event(self, event: Event, span: Span) -> None:
        """Record an event."""
        nvtx.mark(message=event.name_str, color=span.color.value if span.color else None)

    def record_error(self, event: ExceptionEvent, span: Span) -> None:
        """Record an error event."""
        nvtx.mark(message=event.error_message, color=span.color.value if span.color else None)

    def close(self) -> None:
        """Close the NVTX backend."""
        try:
            for range in self.ranges_by_span_id.values():
                nvtx.end_range(range)
        finally:
            self.ranges_by_span_id.clear()



#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#
import json
import os
from io import TextIOWrapper
from typing import Any

from training_telemetry.config import BackendType
from training_telemetry.internal.backends.event_based_backend import EventBasedBackend, RecordWriter
from training_telemetry.verbosity import Verbosity


class FileRecordWriter(RecordWriter):
    """A record writer that writes to a file."""

    def __init__(self, filepath: str):
        self._filepath: str = filepath
        self._file: TextIOWrapper | None = self._open_file()

    def _open_file(self) -> TextIOWrapper:
        parent_dir = os.path.dirname(self._filepath)
        if parent_dir:
            os.makedirs(parent_dir, exist_ok=True)
        return open(self._filepath, "w")

    def serialize(self, record: dict[str, Any]) -> str:
        return json.dumps(record, separators=(",", ":"))

    def write(self, record: str) -> None:
        if not self._file:
            raise RuntimeError("File backend not initialized or already closed")
        self._file.write(record + "\n")
        self._file.flush()

    def close(self) -> None:
        if self._file is not None:
            self._file.close()
            self._file = None


class FileBackend(EventBasedBackend):
    """Backend implementation that writes events and metrics to a JSONfile."""

    def __init__(self, filepath: str):
        self._record_writer = FileRecordWriter(filepath)
        super().__init__(self._record_writer)

    def type(self) -> BackendType:
        return BackendType.FILE

    def verbosity(self) -> Verbosity:
        return Verbosity.INFO



#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#

import os
from abc import ABC, abstractmethod
from threading import Lock
from typing import Any, Optional

from training_telemetry.events import Event, ExceptionEvent
from training_telemetry.internal.backend import Backend
from training_telemetry.internal.backends.record_type import RecordType
from training_telemetry.spans import Span
from training_telemetry.utils import get_default_logger

_logger = get_default_logger()


class RecordWriter(ABC):
    """A record writer is used by the EventBasedBackend to serialize andwrite records to a backend."""

    @abstractmethod
    def serialize(self, record: dict[str, Any]) -> str:
        """Serialize a record, encoded as a dict, to a raw string for writing in write."""

    @abstractmethod
    def write(self, record: str) -> None:
        """Write a serialized record to the backend, e.g. a python logger, or a file."""

    @abstractmethod
    def close(self) -> None:
        """Close the writer, e.g. a file."""


class EventBasedBackend(Backend):
    """
    An abstract backend base class that records events in order to represent spans.
    Callers must supply a RecordWriter, and implement the backent type and verbosity methods.
    """

    def __init__(self, record_writer: RecordWriter):
        self._record_count: int = 0
        self.pid = os.getpid()
        self._lock: Lock = Lock()
        self._last_start_span: Optional[Span] = None
        self._record_writer = record_writer
        self._closed = False

    def _write_record(self, record: dict[str, Any]) -> None:
        """Write a serialized record to the backend, e.g. a python logger, or a file."""
        if self._closed:
            raise RuntimeError("Backend is closed")

        try:
            with self._lock:
                self._record_count += 1
                record["count"] = self._record_count
                record["pid"] = self.pid
                record_str = self._record_writer.serialize(record)
                self._record_writer.write(record_str)
        except Exception as e:
            _logger.error(f"Error writing record to backend: {e}")
            raise

    def record_span(self, span: Span, record_type: RecordType) -> None:
        """Record a span."""
        record: dict[str, Any] = {
            "type": record_type.value,
            "id": str(span.id),
            "name": span.name.value,
        }
        if record_type == RecordType.START:
            record["event"] = span.start_event.to_json()
        elif record_type == RecordType.COMPLETE:
            assert span.stop_event
            record["elapsed"] = span.duration.elapsed
            record["start_event"] = span.start_event.to_json()
            record["stop_event"] = span.stop_event.to_json()
        elif record_type == RecordType.STOP:
            assert span.stop_event
            record["elapsed"] = span.duration.elapsed
            record["event"] = span.stop_event.to_json()
        else:
            raise ValueError(f"Invalid type: {record_type}")

        self._write_record(record)

    def record_start(self, span: Span) -> None:
        if self._last_start_span is not None:
            self.record_span(self._last_start_span, RecordType.START)
        self._last_start_span = span

    def record_stop(self, span: Span) -> None:
        if span.stop_event is None:
            raise ValueError(f"Span {span.id} has no stop event")
        if self._last_start_span:
            if self._last_start_span.id == span.id:
                # record a complete span rather than two
                self.record_span(span, RecordType.COMPLETE)
                self._last_start_span = None
                return
            else:
                # record the previous span before recording the stop event
                self.record_span(self._last_start_span, RecordType.START)
                self._last_start_span = None

        # record the stop event
        self.record_span(span, RecordType.STOP)

    def record_event(self, event: Event, span: Span) -> None:
        if self._last_start_span:
            self.record_span(self._last_start_span, RecordType.START)
            self._last_start_span = None
        record = {
            "type": RecordType.EVENT.value,
            "id": str(span.id),
            "name": span.name.value,
            "event": event.to_json(),
        }
        self._write_record(record)

    def record_error(self, event: ExceptionEvent, span: Span) -> None:
        if self._last_start_span:
            self.record_span(self._last_start_span, RecordType.START)
            self._last_start_span = None

        record = {
            "type": RecordType.ERROR.value,
            "id": str(span.id),
            "name": span.name.value,
            "event": event.to_json(),
        }
        self._write_record(record)

    def close(self) -> None:
        if self._closed:
            return

        if self._last_start_span:
            _logger.warning(f"Span {self._last_start_span.id} was not closed")

        with self._lock:
            self._last_start_span = None
            self._record_writer.close()
            self._closed = True



#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#
import json
import logging
import re
from typing import Any

from training_telemetry.config import BackendType
from training_telemetry.internal.backends.event_based_backend import EventBasedBackend, RecordWriter
from training_telemetry.verbosity import Verbosity


class PythonLoggerRecordWriter(RecordWriter):
    """A record writer that writes to a python logger."""

    def __init__(self, logger: logging.Logger):
        self._logger = logger

    def serialize(self, record: dict[str, Any]) -> str:
        return self.format_log(record)

    def format_log(self, event_json: dict[str, Any]) -> str:
        """
        Format the dictionary into a log string. The output string will be in the format of "[key1=value1 | key2=value2 | ...]".
        Nested dictionaries are formatted recursively, and will be in the format of "key1.key2=value where key1 is the top
        level key and key2 is the nested key in the nested dictionary".

        Args:
            event_json (dict): The dictionary to format.

        Limitations:
            - dictionary keys cannot contain dots.
        Returns:
            str: The formatted log string.
        """
        items: list[str] = []
        for k, v in sorted(event_json.items(), reverse=True):
            if "." in k:
                raise ValueError(f"keys cannot contain dots: {k}")

            if isinstance(v, dict):
                items.append(self._format_log_nested(event_json=v, prefix=k))
            else:
                items.append(f"{k}={v}")

        return "[" + " | ".join(items) + "]"

    def _format_log_nested(self, event_json: dict[str, Any], prefix: str) -> str:
        """
        Helper function called by format_log to recursively format nested dictionaries.
        Args:
            event_json (dict): The nesteddictionary to format.
            prefix (str): The prefix is the key of the nested dictionary in the parent dictionary.
                          This function is in called recursively to format nested dictionaries,
                          with the prefix as the parent key.
        """
        items: list[str] = []
        for k, v in sorted(event_json.items(), reverse=True):
            if "." in k:
                raise ValueError(f"keys cannot contain dots: {k}")

            if isinstance(v, dict):
                items.append(self._format_log_nested(event_json=v, prefix=f"{prefix}.{k}"))
            else:
                items.append(f"{prefix}.{k}={v}")

        return " | ".join(items)

    @staticmethod
    def deserialize_log(text: str) -> dict[str, Any]:
        """
        Unformat a log string back into a dictionary by reversing the format_log function. Refer to that function
        for details on the format of the log string.
        """
        # First extract all the characters that are in the top level square brackets.
        pattern = r"\[(.*?)\]"
        matches = re.findall(pattern, text, re.DOTALL)
        if not matches:
            raise ValueError(f"Log string must be enclosed in square brackets: {text}")

        # Then extract the content of the last match (the main log body inserted by format_log)
        # and split on the pipe character but exclude the pipe characters that are inside double quotes.
        # Pattern explanation:
        # - r'\|(?=(?:[^"]*"[^"]*")*[^"]*$)'
        # - \|: Matches the pipe character
        # - (?=...) is a positive lookahead assertion
        # - (?:[^"]*"[^"]*")* matches pairs of quotes and their contents
        # - [^"]*$ ensures we're not inside quotes by checking that there are an even number of quotes after the pipe
        content = matches[-1]
        pattern = r'\|(?=(?:[^"]*"[^"]*")*[^"]*$)'
        items = re.split(pattern, content)
        items = [item.strip() for item in items if item.strip()]

        ret: dict[str, Any] = {}
        for item in items:
            # we know keys won't contain the = or . character, so splitting
            # on the first one is safe even if there a = character in the value
            k, v = item.split("=", 1)
            d = ret
            while "." in k:
                k1, k2 = k.split(".", 1)
                if k1 not in d:
                    d[k1] = {}
                k = k2
                d = d[k1]
            try:
                d[k] = json.loads(v)
            except ValueError:
                d[k] = v
        return ret

    def write(self, record: str) -> None:
        self._logger.info(record)

    def close(self) -> None:
        pass


class PythonLoggerBackend(EventBasedBackend):
    """
    Backend implementation that writes spans and events using a python logger.
    Note that this python logger may be backed by an Open Telemetry logger.
    Events are formatted as key=value pairs with hierarchical keys separated by dots.
    """

    def __init__(self, logger: logging.Logger, verbosity: Verbosity):
        self._logger = logger
        self._verbosity: Verbosity = verbosity
        self._record_writer = PythonLoggerRecordWriter(logger)
        super().__init__(self._record_writer)

    def type(self) -> BackendType:
        return BackendType.LOGGER

    def verbosity(self) -> Verbosity:
        return self._verbosity
